@@ -7,6 +7,7 @@ import qualified Data.ByteString as B
 import qualified Data.Text.Lazy as T
 import Data.Text.Lazy.Encoding (decodeUtf8)
 import qualified Data.Certificate.X509 as X509
+import Data.Certificate.X509.Cert as Cert
 import Data.Certificate.KeyRSA as KeyRSA
 import Data.Certificate.KeyDSA as KeyDSA
 import Data.List (find)
@@ -16,16 +17,24 @@ import Control.Monad
 import Control.Applicative ((<$>))
 import Data.Maybe
 import System.Exit
-import System.Certificate.X509 as SysCert
+import System.Certificate.X509
+import Data.CertificateStore
 
 -- for signing/verifying certificate
 import qualified Crypto.Hash.SHA1 as SHA1
+import qualified Crypto.Hash.SHA224 as SHA224
+import qualified Crypto.Hash.SHA256 as SHA256
+import qualified Crypto.Hash.SHA512 as SHA512
 import qualified Crypto.Hash.MD2 as MD2
 import qualified Crypto.Hash.MD5 as MD5
-import qualified Crypto.Cipher.RSA as RSA
-import qualified Crypto.Cipher.DSA as DSA
+import qualified Crypto.PubKey.HashDescr as HD
+import qualified Crypto.PubKey.RSA as RSA
+import qualified Crypto.PubKey.RSA.PKCS15 as RSA
+import qualified Crypto.PubKey.DSA as DSA
 
-import Data.ASN1.DER (decodeASN1Stream, ASN1(..), ASN1ConstructionType(..))
+import Data.ASN1.Encoding
+import Data.ASN1.BinaryEncoding
+import Data.ASN1.Stream
 import Data.ASN1.BitArray
 import Text.Printf
 import Numeric
@@ -36,7 +45,10 @@ hexdump bs = concatMap hex $ L.unpack bs
 		| n > 0xa   = showHex n ""
 		| otherwise = "0" ++ showHex n ""
 
-showDN dn = mapM_ (\(oid, (_,t)) -> putStrLn ("  " ++ show oid ++ ": " ++ t)) dn
+hexdump' :: B.ByteString -> String
+hexdump' = hexdump . L.fromChunks . (:[])
+
+showDN (X509.DistinguishedName dn) = mapM_ (\(oid, (_,t)) -> putStrLn ("  " ++ show oid ++ ": " ++ t)) dn
 
 showExts es = do
 	mapM_ showExt es
@@ -79,7 +91,7 @@ showCert (X509.X509 cert _ _ sigalg sigbits) = do
 		X509.PubKeyUnknown oid ws -> do
 			printf "public key unknown: %s\n" (show oid)
 			printf "  raw bytes: %s\n" (show ws)
-			case decodeASN1Stream $ L.pack ws of
+			case decodeASN1 BER $ L.pack ws of
 				Left err -> printf "  asn1 decoding failed: %s\n" (show err)
 				Right l  -> printf "  asn1 decoding:\n" >> showASN1 4 l
 		pk                        ->
@@ -108,13 +120,13 @@ showRSAKey (pubkey,privkey) = unlines
 
 showDSAKey :: (DSA.PublicKey,DSA.PrivateKey) -> String
 showDSAKey (pubkey,privkey) = unlines
-	[ "priv     " ++ (show $ DSA.private_x privkey)
-	, "pub:     " ++ (show $ DSA.public_y pubkey)
-	, "p:       " ++ (show p)
-	, "q:       " ++ (show g)
-	, "g:       " ++ (show q)
+	[ "priv     " ++ (printf "%x" $ DSA.private_x privkey)
+	, "pub:     " ++ (printf "%x" $ DSA.public_y pubkey)
+	, "p:       " ++ (printf "%x" p)
+	, "q:       " ++ (printf "%x" g)
+	, "g:       " ++ (printf "%x" q)
 	]
-    where (p,g,q) = DSA.private_params privkey
+    where (p,q,g) = DSA.private_params privkey
 
 showASN1 :: Int -> [ASN1] -> IO ()
 showASN1 at = prettyPrint at where
@@ -160,16 +172,24 @@ parsePEMCert = either (const []) (rights . map getCert) . pemParseBS
 
 processCert opts (cert, x509) = do
 	when (raw opts) $ putStrLn $ hexdump $ L.fromChunks [cert]
-	when (asn1 opts) $ case decodeASN1Stream $ L.fromChunks [cert] of
+	when (asn1 opts) $ case decodeASN1' BER cert of
 		Left err   -> error ("decoding ASN1 failed: " ++ show err)
 		Right asn1 -> showASN1 0 asn1
 
 	when (text opts || not (or [asn1 opts,raw opts])) $ showCert x509
-	when (verify opts) $ verifyCert x509
+	when (hash opts) $ hashCert x509
+	when (verify opts) $ getSystemCertificateStore >>= flip verifyCert x509
 	where
-		verifyCert x509@(X509.X509 cert _ _ sigalg sig) = do
-			sysx509 <- SysCert.findCertificate (matchsysX509 cert)
-			case sysx509 of
+		hashCert x509@(X509.X509 cert _ _ _ _) = do
+			putStrLn ("subject(MD5):  " ++ hexdump' (X509.hashDN_old subject))
+			putStrLn ("issuer(MD5):   " ++ hexdump' (X509.hashDN_old issuer))
+			putStrLn ("subject(SHA1): " ++ hexdump' (X509.hashDN subject))
+			putStrLn ("issuer(SHA1):  " ++ hexdump' (X509.hashDN issuer))
+			where
+				subject    = X509.certSubjectDN cert
+				issuer     = X509.certIssuerDN cert
+		verifyCert store x509@(X509.X509 cert _ _ sigalg sig) = do
+			case findCertificate (X509.certIssuerDN cert) store of
 				Nothing                        -> putStrLn "couldn't find signing certificate"
 				Just (X509.X509 syscert _ _ _ _) -> do
 					verifyAlg (B.concat $ L.toChunks $ X509.getSigningData x509)
@@ -177,16 +197,21 @@ processCert opts (cert, x509) = do
 					          sigalg
 					          (X509.certPubKey syscert)
 
-		rsaVerify h hdesc pk a b = either (Left . show) (Right) $ RSA.verify h hdesc pk a b
+		rsaVerify hdesc pk a b = Right $ RSA.verify hdesc pk a b
 
 		verifyF (X509.SignatureALG hash X509.PubKeyALG_RSA) (X509.PubKeyRSA rsak) =
-			let (f, asn1) = case hash of
-				X509.HashMD2  -> (MD2.hash, "\x30\x20\x30\x0c\x06\x08\x2a\x86\x48\x86\xf7\x0d\x02\x05\x05\x00\x02\x10")
-				X509.HashMD5  -> (MD5.hash, "\x30\x20\x30\x0c\x06\x08\x2a\x86\x48\x86\xf7\x0d\x02\x05\x05\x00\x04\x10")
-				X509.HashSHA1 -> (SHA1.hash, "\x30\x21\x30\x09\x06\x05\x2b\x0e\x03\x02\x1a\x05\x00\x04\x14")
-				_             -> error ("unsupported hash in RSA: " ++ show hash)
+			let hdesc = case hash of
+				-- "ASN.1 DER X algorithm designator prefix"
+				X509.HashMD2    -> HD.hashDescrMD2
+				X509.HashMD5    -> HD.hashDescrMD5
+				X509.HashSHA1   -> HD.hashDescrSHA1
+				X509.HashSHA224 -> HD.hashDescrSHA224
+				X509.HashSHA256 -> HD.hashDescrSHA256
+				X509.HashSHA384 -> HD.hashDescrSHA384
+				X509.HashSHA512 -> HD.hashDescrSHA512
+				_               -> error ("unsupported hash in RSA: " ++ show hash)
 				in
-			rsaVerify f asn1 rsak
+			rsaVerify hdesc rsak
 
 		verifyF (X509.SignatureALG _ X509.PubKeyALG_DSA) (X509.PubKeyDSA dsak) =
 			(\_ _ -> Left "unimplemented DSA checking")
@@ -234,6 +259,7 @@ data CertMainOpts =
 		, text   :: Bool
 		, raw    :: Bool
 		, verify :: Bool
+		, hash   :: Bool
 		}
 	| Key
 		{ files :: [FilePath]
@@ -246,6 +272,7 @@ x509Opts = X509
 	, text   = def
 	, raw    = def
 	, verify = def
+	, hash   = def
 	} &= help "x509 certificate related commands"
 
 keyOpts = Key
